;----------------------------------------------------------------------------
; MBus Slave receive routines for TI-83
; This is a port of the ZShell 4.0 version.
; Functionality and calling procedure is identical to ZShell version.
; Porting performed by Vasantha Crabb <_merlin@operamail.com>
;----------------------------------------------------------------------------

;----------------------------------------------------------------------------
; MBus Slave receive routines for TI-85 & ZShell 4.0
; Version 0.99 (public beta) 14.1.1998
; Copyright (C) 1998 Osma Suominen <ozone@clinet.fi>
;----------------------------------------------------------------------------
;
;  This library is free software; you can redistribute it and/or
;  modify it under the terms of the GNU Library General Public
;  License as published by the Free Software Foundation; either
;  version 2 of the License, or (at your option) any later version.
;
;  This library is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;  Library General Public License for more details.
;
;  You should have received a copy of the GNU Library General Public
;  License along with this library; if not, write to the Free
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;----------------------------------------------------------------------------
;
; Usage:
; - Set the address of this calculator into I2CAddress
; - CALL_(SI2CLoad) once to load the interrupt handler
;   and reset the link port.
; - Occasionally check whether I2CInputLength is > 0, and if it is, read
;   any data in the buffer and set I2CInputLength back to 0.
;   I2CInputLength shows how many bytes are in the input buffer.
;   It can also be used to disable receiving of packets for a short time,
;   by incrementing it, and then decrementing it back to its original value
;   when you want to enable receiving. This kind of disabling only inhibits
;   the receiving of whole packets, the routines still acknowledge the
;   calculator's own address byte to show that it is still connected.
; - Set IM 1 before returning to ZShell (and clear graph mem if you like)
;
; Uses these memory regions (in APD buffer):
;
; $8228-$8281     unused
; $8282-$83xx     the interrupt routine
; $8400-$8500     full of $82
; $8501-$8527     unused (could be used for variables - see below)
;
; Variables:      (these must be defined by the application)
;
; I2CAddress      Own I2C address (should be even, xxxxxxx0)           (byte)
; I2CInputBuffer  Start of input buffer                           (128 bytes)
; I2CInputLength  Shows how many bytes have been stored in the buffer  (byte)
;
; I2CCStart       Start counter                                        (word)
; I2CCStop        Stop counter                                         (word)
; I2CCMiss        Missed Start counter                                 (word)
; I2CCTimeout     Timeout counter                                      (word)
; I2CCChecksum    Checksum error counter                               (word)
;
; The data format in the input buffer:
;
; Offset                Content
; 0                     The sender's address (xxxxxxx1 for rcvd general calls)
; 1                     Data length (n)
; 2                     Protocol id
; 3 ->                  Data (n-1 bytes) (One less than given to MI2C in B)

#define IntBase $8282 - SI2CIntRStart

SI2CLoad:
    xor     a                   ; A=0
    ld      (I2CInputLength),a  ; There's no input (yet)
    im      1                   ; Disable other interrupt routines
    ld      c,$82
    ld      hl,$8400
    ld      b,a                 ; 0 => 256 times
SI2CLoadLoop:
    ld      (hl),c              ; Load $82 into ($8400)
    inc     hl
    djnz    SI2CLoadLoop
    ld      (hl),c
    ld      hl,SI2CIntRStart
    ld      de,$8282
    ld      bc,SI2CIntREnd-SI2CIntRStart
    ldir                        ; Load the interrupt routine into graph mem
    ld      a,$D0
    out     (BPORT),a           ; Reset link port
    ld      a,$84               ; Set the interrupt vector to point to $84xx
    ld      i,a
    im      2                   ; Enable interrupt routine
    ei
    ret

; The actual interrupt routine:
; Software protocol layer of slave routines

SI2CIntRStart:
    exx                             ; Swap in alternate register set
    ex      af,af'

    ld      a,$D0
    out     (BPORT),a

; Detect Start
    in      a,(BPORT)
    cp      24
    jr      c,SI2CMissedStart
    jp      nz,SI2CIntEnd+IntBase

SI2CDetLoop:                        ; Waitfor SCL=0
    in      a,(BPORT)
    rra
    rra
    rra
    jr      c,SI2CMissedStart       ; If SDA returns to 1, we've missed a Start
    rra                             ; CY=SCL
    jr      c,SI2CDetLoop           ; Waitfor SCL=0
                                    ; Start has been detected
    ld      a,$D2                   ; SCL=0 SDA=1
    out     (BPORT),a               ; Pull down SCL to halt transfer
SI2CStart:
    ld      hl,(I2CCStart)          ; Increase Start counter
    inc     hl
    ld      (I2CCStart),hl
    ld      c,0                     ; Receive an address
    ld      a,(I2CAddress)
    ld      e,a                     ; Initialize checksum counter with own addr
    call    SI2CRecv+IntBase        ; Read destination address
    ld      a,(I2CInputLength)      ; Check if input buffer is empty
    or      a                       ; Same as cp 0
    jr      nz,SI2CDetStop          ; Do not Ack the next bytes if it isn't
    ld      hl,I2CInputBuffer
                                    ; Now receive the rest of the packet
    ld      b,2                     ; Sender's address and length, 2 bytes
    inc     c                       ; Receive data (ack everything)
SI2CRecvHeader:
    call    SI2CRecv+IntBase
    ld      (hl),d                  ; Store received byte
    inc     hl
    djnz    SI2CRecvHeader
    ld      b,d                     ; Read <length> bytes
SI2CRecvData:
    call    SI2CRecv+IntBase
    ld      (hl),d                  ; Store received byte
    inc     hl
    djnz    SI2CRecvData
    inc     c                       ; Now receive the checksum byte
    call    SI2CRecv+IntBase
    ld      a,(I2CInputBuffer+1)    ; No. of data bytes in the packet
    add     a,2                     ; +2: the src addr and data length bytes
    ld      (I2CInputLength),a      ; Packet accepted
    jr      SI2CDetStop             ; No more interesting bytes should come..

SI2CDetStCount:
    dec     c
    jr      nz,SI2CDetStop2
    jr      SI2CCountTimeout

SI2CMissedStart:
    ld      hl,(I2CCMiss)           ; Increase Missed Start counter
    inc     hl
    ld      (I2CCMiss),hl

SI2CDetStop:
    ld      a,$D0                   ; SDA=1 SCL=1
    out     (BPORT),a               ; Release both SCL and SDA
SI2CDetStop1:
    ld      c,0                     ; Timeout counter, 0 -> 256
SI2CDetStop2:                       ; Waitfor SCL=1 and SDA=0
    in      a,(BPORT)
    cp      24
    jr      c,SI2CDetStop1          ; If SCL=0, new loop, reset counter
    jr      nz,SI2CDetStCount       ; If SDA=SCL=1, count timeout
SI2CDetStop3:                       ; Waitfor SCL=1 and SDA=1
    in      a,(BPORT)
    cp      24
    jr      c,SI2CDetStop1          ; If SCL=0, new loop, reset counter
    jr      z,SI2CDetStop3

                                    ; Stop has now been detected
SI2CCountStop:
    ld      hl,(I2CCStop)           ; Increase Stop counter
    inc     hl
    ld      (I2CCStop),hl

SI2CIntEnd:
    ex      af,af'
    exx
    jp      $38

SI2CDetStopPop:
    pop     bc
    pop     af
    jr      SI2CDetStop

SI2CTimeout:
    ld      a,$D0                   ; Reset link port (because of Ack)
    out     (BPORT),a
    pop     bc
    pop     af                      ; Return address
SI2CCountTimeout:
    ld      hl,(I2CCTimeout)        ; Increase Timeout counter
    inc     hl
    ld      (I2CCTimeout),hl
    jr      SI2CIntEnd

SI2CChecksumError:
    ld      hl,(I2CCChecksum)       ; Increase Checksum Error counter
    inc     hl
    ld      (I2CCChecksum),hl
    jr      SI2CDetStopPop

; Hardware protocol layer of slave routines
; Pulls down SCL between bits to synchronize speed with master
; Receive a byte of data and ack/nack it according to the following rules:
; If bit 0,c = 1 then always ack
; If bit 1,c = 1 then ack if received byte = E
; If bit 1,c = 0 then ack if received byte = E or = 0 (general call)
; C must have bits 2-7 reset (0)
; That means C sets the mode:
; C=0 address mode
; C=1 data mode
; C=2 checksum mode
; Received byte is returned in D and checksum is counted in E

SI2CRecv:
    ld      a,$D0                   ; SCL=1 SDA=1
    out     (BPORT),a               ; Ready to receive a byte - release SCL
    push    bc
    ld      b,8                     ; 1 byte
SI2CRecvBit:
    in      a,(BPORT)
    cp      24
    jr      c,SI2CRecvBit           ; waitfor SCL=1
    jr      z,SI2CRcv_0

;!!
    ld      c,0                     ; Timeout counter - 100 should be reasonable
                                    ; (Can receive when sender uses speed 0)
SI2CRcv_1:
    in      a,(BPORT)               ; xxxxxxCD 1
    dec     c
    jr      z,SI2CTimeout           ; A timeout...
    sub     24
    jr      c,SI2CRcvTail           ; Bit received OK, CY=bit=1
    jr      nz,SI2CRcv_1
                                    ; Start!
    pop     bc
    pop     af                      ; Return address

; Support for repeated START conditions can be enabled by uncommenting
; the following line and commenting out the next one. However, they're
; currently not used in the Master routines and may potentially cause
; trouble, so they should be disabled.

;        jp SI2CDetLoop+IntBase ; Start over, first wait for SCL=0
    jr      SI2CDetStop

SI2CRcv_0:
    in      a,(BPORT)
    add     a,232                   ; Same as sub 24 but CY behaviour is inverted
    jr      nc,SI2CRcvTail          ; Bit received OK, CY=bit=0
    jr      z,SI2CRcv_0
                                    ; Stop!
    pop     bc
    pop     af                      ; return address
    jr      SI2CCountStop

SI2CRcvTail:                        ; Received bit in CY
    rl      d                       ; D: xxxxxxxD
    djnz    SI2CRecvBit
    ld      a,$D2                   ; SCL=0 SDA=1
    out     (BPORT),a               ; Pull down SCL to halt transfer

                                    ; Received byte now in D
    pop     bc                      ; B=old counter, C=flags
    push    bc
                                    ; Now determine if we'll ack the byte
    rr      c
    jr      c,SI2CAck               ; Ack every byte
    ld      a,d
    cp      e
    jr      z,SI2CAck               ; Ack own address or matching checksum
    rr      c
    jr      c,SI2CChecksumError     ; Checksums don't match! nack
    or      a
    jr      nz,SI2CDetStopPop       ; Packet not for this calc, nack
                                    ; General call (addr=0), ack
    ld      e,a                     ; Init checksum with general call address 0
SI2CAck:
    ld      a,$D1                   ; SDA=0 SCL=1
    out     (BPORT),a               ; Release SCL - ready to ack
SI2CAck1:
    in      a,(BPORT)
    rra
    rra
    rra
    rra
    jr      nc,SI2CAck1             ; Waitfor SCL=1
    ld      c,0                     ; Timeout counter - 120 should be reasonable
SI2CAck2:
    in      a,(BPORT)
    rra
    rra
    dec     c
    jr      z,SI2CTimeout
    rra
    rra
    jr      c,SI2CAck2              ; Waitfor SCL=0
    ld      a,$D2                   ; SDA=1 SCL=0
    out     (BPORT),a               ; End ack bit, pull down SCL to halt transfer
    pop     bc

    xor     a
    cp      c
    ret     z                       ; Initialize E with own addr if c=0

    ld      a,e                     ; Continue counting the checksum
    xor     d
    ld      e,a                     ; Save the checksum in E
    ret

SI2CIntREnd:                        ; End of interrupt routine
